"""game.py - a context for running a game session or level

This file is part of Gummworld2.

Gummworld2 is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Gummworld2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with Gummworld2.  If not, see <http://www.gnu.org/licenses/>.
"""

import pygame
from pygame.locals import (
    Color, ACTIVEEVENT,
    K_ESCAPE, K_LEFT, K_RIGHT, K_UP, K_DOWN, K_SPACE,
    K_w, K_a, K_s, K_d,
)

import gummworld2
from gummworld2 import State, BasicMapRenderer, Vec2d, Statf
from gummworld2 import model, toolkit

import settings, terrain, player, rules


class Game(gummworld2.Engine):
    
    def __init__(self):
        
        window_title = 'Gummcraft - Space,Up=Jump | Tab=Tool | Mouse1=Dig | Mouse2=Build | Shift=Sprint'
        
        State.speed = settings.player_speed
        self.player = player.Player((0,0))
        
        gummworld2.Engine.__init__(self,
            caption=window_title,
            resolution=settings.resolution,
            camera_target=self.player,
            map=State.map,
            update_speed=settings.update_speed,
            frame_speed=settings.frame_speed,
        )
        State.clock.paused_callback = self.paused
        self.player.rect.centerx = settings.resolution[0]/2
        self.player.rect.bottom = State.map.SEA_LEVEL - 200
        self.player.position = self.player.rect.center
        State.camera.init_position(self.player.position)
        
        State.renderer = self.renderer = BasicMapRenderer(self.map)
        
        self.movex = 0
        self.sprint = 0
        self.mouse_down_1 = None
        self.mouse_down_3 = None
        
        toolkit.make_hud()
        next_pos = State.hud.next_pos
        State.hud.add('playerpos', Statf(next_pos(),
            'Player %s', callback=lambda:str(self.player.rect.center),
            interval=0.2))
        State.hud.add('tool', Statf(next_pos(),
            'Using %s', callback=lambda:str(self.player.tool.name),
            interval=0.2))
        
        State.game = self
        self.do_focus = False
    
    def update(self, dt):
        if pygame.KMOD_SHIFT & pygame.key.get_mods():
            self.sprint = 2 # speed * 2
        else:
            self.sprint = 1 # speed * 1
        rules.update(dt)
        State.camera.update()
        State.renderer.set_rect(center=State.camera.rect.center)
        State.hud.update(dt)
    
    def draw(self, interp):
        State.screen.clear()
        State.renderer.draw_tiles()
        self.draw_player()
        State.hud.draw()
        State.screen.flip()
    
    def draw_player(self):
        # playerpos - cameratopleft = undo interpolation
        camera = State.camera
        surf = camera.view.surface
        player = self.player
        ## de-interpolate the camera target, else it jitters.
        ## Note to self: add this to Camera. It comes up often enough.
        cx,cy = camera.rect.topleft
        px,py = Vec2d(player.rect.center) - camera.rect.center
        surf.blit(player.image, player.rect.move(-(cx+px),-(cy+py)))
        
        # draw mining tool hitbox
        if player.tool.draw_hitbox:
            r = player.tool.rect
            px,py = Vec2d(r.center) - camera.rect.center
            pygame.draw.rect(surf, Color('white'), r.move(-(cx+px),-(cy+py)), 1)
    
    def on_key_down(self, unicode, key, mod):
        keys = settings.KEYS
        if key in keys['left']:    self.movex += -State.speed
        elif key in keys['right']: self.movex +=  State.speed
        elif key in keys['jump']: self.player.jump()
        elif key in keys['swaptool']: self.player.next_tool()
        elif key == K_ESCAPE: self.stop()
    
    def on_key_up(self, key, mod):
        keys = settings.KEYS
        if key in keys['left']:    self.movex -= -State.speed
        elif key in keys['right']: self.movex -=  State.speed
    
    def on_mouse_button_down(self, pos, button):
        if button == 1:
            self.mouse_down_1 = pos
        elif button == 3:
            self.mouse_down_3 = pos
    
    def on_mouse_button_up(self, pos, button):
        if button == 1:
            self.mouse_down_1 = None
        elif button == 3:
            self.mouse_down_3 = None
    
    def on_mouse_motion(self, pos, rel, buttons):
        if buttons[0] and self.mouse_down_1 is not None:
            self.mouse_down_1 = pos
        elif buttons[2] and self.mouse_down_3 is not None:
            self.mouse_down_3 = pos
    
    def on_active_event(self, gain, state):
        if gain == 0:
            # Mouse exited pygame window; auto-pause
            if self.do_focus:
                State.clock.pause()
            else:
                # This works around some kind of weird behavior when the
                # window first opens and the mouse position is polled.
                # Sometimes pygame thinks the mouse has left the window
                # area, or started outside the window area or something,
                # firing an ACTIVEEVENT with gain=0. Then you must drag
                # the mouse out and back in to receive gain=1.
                self.do_focus = True
        else:
            # Mouse entered pygame window; auto-unpause
            State.clock.resume()
    
    def paused(self):
        e = pygame.event.wait()
        if e.type == ACTIVEEVENT:
            if e.gain == 1:
                # Mouse entered pygame window; reset some stuff
                # Need to check if pressed, and set the value appropriately.
                keys = pygame.key.get_pressed()
                self.movex = 0
                for i in settings.KEYS['left']:
                    if keys[i]:
                        self.movex -= State.speed
                for i in settings.KEYS['right']:
                    if keys[i]:
                        self.movex += State.speed
                self.mouse_down_1 = False
                self.mouse_down_3 = False
                State.clock.resume()
                pygame.event.clear()
    
    def on_quit(self):
        self.stop()
    
    def stop(self):
        self.pop()
