import pygame

from gummworld2 import Vec2d

import _types


class Node(object):
    """A node of substance in the world
    
    size is the size of the node in pixels.
    
    anchor is the attribute of the rect used when setting the "position"
    attribute.
    
    images is undefined. It must be created by the subclass as a list.
    """
    size = 16,16
    anchor = 'center'
    
    deformable = True
    max_health = 1000.0
    time_till_reset = 10.0
    
    images = None
    
    def __init__(self, position, image_files=[], start_image=0):
        self.image = None
        self.rect = None
        self._position = Vec2d(position)
        self._start_image = start_image
        
        if not self.images:
            self._load_images(image_files)
        
        self.image = self.images[self._start_image]
        self.rect = self.image.get_rect()
        self.position = self.position
        self.image_idx = 0
        
        self.health = self.max_health
        self.time_since_hit = 0.0
        
        self.dirty = False
    
    def _set_position(self, val):
        self._position[:] = val
        x = int(round(val[0]))
        y = int(round(val[1]))
        setattr(self.rect, self.anchor, (x,y))
    def _get_position(self):
        return self._position[:]
    position = property(_get_position, _set_position,
        "position is stored as a float vector. Setting position will also" +
        " update the pygame rect attribute according to the class or" +
        " instance attribute 'anchor'.")
    
    def _load_images(self, image_files):
        for name in image_files:
            ## this would typically load images from files
            im = pygame.Surface(self.size)
            im.fill(pygame.Color(name))
            self.images.append(im)
    
    def _color_outline(self, color):
        ## this would typically load images from files
        x,y,w,h = self.rect
        pygame.draw.line(self.image, color, (w/3,h-1), (w,h-1), 1)
        pygame.draw.line(self.image, color, (w-1,h), (w-1,h/3), 1)
    
    def heal(self, dt):
        self.time_since_hit -= dt
        if self.time_since_hit <= 0.0:
            self.health = self.max_health
            self._calc_image()
    
    def _calc_image(self):
        # calculate the image to display based on health
        n = len(self.images)
        pct = n - int(round(n * self.health / self.max_health))
        if pct != self.image_idx:
            try:
                self.image = self.images[pct]
                self.dirty = True  # signal the caller the map is dirty
                self.image_idx = pct
            except:
                pass
    
    def hit(self, damage):
        if not self.deformable:
            return
        
        # decrement health and refresh time to reset
        self.health -= damage
        self.time_since_hit = self.time_till_reset
        
        # defeated
        if self.health < 0.0:
            self.health = 0.0
            self.time_since_hit = 0.0
            return
        
        self._calc_image()
